package provide ade_lib 1.0

set CIMS_USR /usr

proc create_aduser {dn upn sam pw {dname {}} {gname {}} {spn {}} {gecos {}}} {
    new_object $dn;
    sof userPrincipalName $upn;
    sof sAMAccountName $sam;
    sof objectClass user;
    sof userAccountControl 514;
    save_object;
    append samAtDomain "$sam@" [domain_from_dn $dn]
    sup $samAtDomain $pw;
    sof userAccountControl 512;
    save_object;
    slo $dn
    if {[string length $dname]} {
        sof displayName $dname
    }
    if {[string length $gname]} {
        sof givenName $gname
    }
    if {[string length $spn]} {
        sof servicePrincipalName $spn
    }
    if {[string length $gecos]} {
        sof gecos $gecos
    }
    svo
}

proc get_user_groups {args} {
    push
    set usage {wrong # args: should be [-dn] [-z] <user dn|user upn>} 
    set getdn [getopt args "-dn"]
    set zone_enable [getopt args "-z"]
    if { [llength $args] != 1 } {
        error $usage
    }
    lassign $args aduser
    set group_dn {}
    set groups {}
    
    if { $zone_enable } {
        foreach g [get_zone_groups] {
            dict set group_dn [string tolower [principal_to_dn $g]] 1
        }
    }

    slo [GetDN $aduser]
    foreach g [gof memberOf] {
        if { $zone_enable && ![dict exists $group_dn [string tolower $g]] } {
            continue
        }
        lappend groups [expr { $getdn ? $g : [dn_to_principal $g] }]
    }
    pop
    return $groups
}

proc create_adgroup {dn sam gtype} {
    new_object $dn;
    sof sAMAccountName $sam;
    sof objectClass group;
    switch $gtype {
        "global" {
                sof groupType [expr 0x80000002]
        }
        universal {
                sof groupType [expr 0x80000008]
        }
        local {
                sof groupType [expr 0x80000004]
        }
        default {
            error "invalid group type"
        }
    }

    save_object;
}

proc create_user {ad uname uid gid gecos home shell role} {

    set ztype [gzf type]
    if {$uname ne "-"
            || $uid ne "-"
            || $gid ne "-"
            || $gecos ne "-"
            || $home ne "-"
            || $shell ne "-"} {
        new_zone_user $ad;
        szuf uname $uname;
        szuf uid $uid;
        szuf gid $gid;
        # you cannot set the gecos field on classsic zones, so warn user and continue
        catch { szuf gecos $gecos;}
        szuf home $home;
        szuf shell $shell;
        if {[string match classic* $ztype]} {
            szuf enabled $role
        }
        save_zone_user;
    }
    if {[string match classic* $ztype] eq 0} {
        if {$role ne "-"} {
            if {[string first "+" $ad] == -1} {
                new_role_assignment $ad;
                set_role_assignment_field role $role;
                save_role_assignment;
            }
        }
    }
}


#
#  Create Local user profile in the selected zone.
#    - uname is required argument, the rest are optional
#    - Specify "-" to not set any value for the attribute in the profile. 
#    - Empty value can be specified to gecos arg. 
# 
proc create_local_user_profile {uname uid gid gecos home shell profileflag role} {
    set ztype [gzf type];
    if {[string match tree $ztype] eq 0} {
        error "zone type $ztype is not supported"
    }
    if {[string length $uname] eq 0
            || $uname eq "-"} {
        error "uname can not be empty"
    }
    new_local_user_profile $uname;
    if {$uid ne "-"
            || $gid ne "-"
            || $gecos ne "-"
            || $home ne "-"
            || $shell ne "-"
            || $profileflag ne "-"} {    
        slupf uid $uid;
        slupf gid $gid;            
        slupf gecos $gecos;
        slupf home $home;
        slupf shell $shell;
        slupf profileflag $profileflag;     
    }
    save_local_user_profile;
    if {$role ne "-"} {
        new_role_assignment [append uname "@localhost"];
        set_role_assignment_field role $role;
        save_role_assignment;         
    }
}

proc lremove {listVariable value} {
    upvar 1 $listVariable var
    set var [lsearch -not -inline -all -nocase $var $value]
}

# merge the lists into one and lsort -unique it.
proc lmerge {args} {
    if {[llength $args] < 1} {
	error "args: lists to be merged and sorted"
    }
    return [lsort -unique [concat {*}$args]]
}

# args: upn/dn
proc GetDN { object } {
    if {[regexp -nocase {(cn|ou|dc)=.+} $object]} {
        return $object
    }
    return [principal_to_dn $object]
}

# args: upn/udn, gpn/gdn.
proc remove_user_from_group {user group} {
    push
    set gdn [GetDN $group]
    set udn [GetDN $user]
    set udom [domain_from_dn $udn]
    set uforest [get_bind_info $udom forest]
    set gdom [domain_from_dn $gdn]
    set gforest [get_bind_info $gdom forest]
    if {$uforest != $gforest} {
         slo $udn
         set sid [gof sid]
         set udn "<SID="
         append udn $sid ">"
    }

    if {[catch {remove_object_value $gdn member $udn} err]} {
        pop
        error "$err"
    }

    pop
}

# args: upn/udn, gpn/gdn.
proc add_user_to_group {user group} {
    push
    set gdn [GetDN $group]
    set udn [GetDN $user]
    if { [string tolower $gdn] == [string tolower $udn] } {
        error "Cannot add $gdn to itself."
    }
    set udom [domain_from_dn $udn]
    set uforest [get_bind_info $udom forest]
    set gdom [domain_from_dn $gdn]
    set gforest [get_bind_info $gdom forest]
    if {$uforest != $gforest} {
        slo $udn
        set sid [gof sid]
        set udn "<SID="
        append udn $sid ">"
    }

    if {[catch {add_object_value $gdn member $udn} err]} {
        pop
        error "$err"
    }
    pop
}

# from/to, in the format of "yr-mon-day hr:min"
proc create_assignment {adg role {from {}} {to {}} {description {}}} {
    new_role_assignment $adg;
    set_role_assignment_field role $role;
    if {[string length $from]} {
        sraf from [clock scan $from]
    }
    if {[string length $to]} {
        sraf to [clock scan $to]
    }
    if {[string length $description]} {
        sraf description $description
    }
    save_role_assignment;
}

proc create_group {adg name gid {req {}}} {
    new_zone_group $adg;
    szgf name $name;
    szgf gid $gid;
    if {[string length $req]} {
        szgf required $req
    }
    save_zone_group;
}

#
#  Create Local Group profile in the selected zone.
#    - gname is required argument, the rest are optional.
#    - Specify "-" to not set any value for the attribute in the profile. 
#    - members is a comma separated list.
#    - Empty value can be specified to members arg. 
#
proc create_local_group_profile {gname gid members profileflag} {
    set ztype [gzf type];
    if {[string match tree $ztype] eq 0} {
        error "zone type $ztype is not supported"
    }
    if {[string length $gname] eq 0
            || $gname eq "-" } {
        error "gname can not be empty"
    }
    new_local_group_profile $gname;
    slgpf gid $gid;
    slgpf member $members;
    slgpf profileflag $profileflag
    save_local_group_profile;
}

#args: nismapname, entries(the list of nismap entries, in the format as 
# the result of get_nis_map)
proc create_nismap {nm entries} {
    new_nis_map $nm 
    save_nis_map 
    select_nis_map $nm 
    foreach entr $entries {
        set nm_key [lindex $entr 0]
        set nm_idx [string last ":" $nm_key ]
        add_map_entry [string range $nm_key 0 [ expr $nm_idx-1 ] ] [ lindex $entr 1 ]
    }
    save_nis_map
}


# Precreate a computer, create an zone computer or machine zone
# It will also create the AD computer object if necessary
# Similar to adjoin --precreate
#
# precreate_computer samaccount@domain ... 
# precreate_computer samaccount@domain ?-ad? ?-scp? ?-czone? ?-all? ?-container rdn? ?-dnsname dnsname? ?-trustee upn1? ?-trustee upn2? ?-stype spn1? ?-stype spn2? ?-enctype enc1? ?-enctype enc2? ?-licensetype licensetype? ...
#
# samaccount@domain contains both the computer name and the domain to join
# samaccount should be form of <computer>$, for example, the first argument can be paul-pc$pual.test
# Scp or computer zone will be created in current select zone, so need select_zone first
#
# computer:     computer account name
# Optional Arguments:
# -ad           (optional) Create AD computer object
#                          It will not create AD computer object if the computer object exists
# -scp          (optional) Precreate create zone computer(extension object)
#                          It will create AD computer object if not present
#                          If option specified, ADEdit must be bound to a global catalog domain controller
#                          using the bind command with the -gc option.
#                          To skip the global catalog dc, use the -nogc option. But make sure you understand
#                          the consequence of not doing a global catalog search.                
# -czone        (optional) Create computer zone
# -all          (optional) Create all above
#                          In addition, If no action specified, create ad and scp by default
# -nogc         (optional) It will skip the global catalog dc when creating scp (with the option -scp)
#                          If option specified, it just tries to do a domain search instead of gc search.
#                          Make sure you understand the consequence of not doing a global search, i.e.
#                          adjoin may fail if there are multiple scp of the same computer in the same forest.
# -container <rdn>
#               (optional) Subtree to create AD object
#                          This option takes effect only when a new AD computer object need to be created
#                          If empty, default to CN=Computers,DC=<domain>
# -dnsname <dnsname>
#               (optional) specified dns name
#                          If empty, derive the dnsname from samaccount name and domain
# -trustee <upn>
#               (optional) Arbitrary trustees who can use the precreated computers
#                          Note this option can be set multiple times
# -stype <spn>
#               (optional) Specified servicePrincipalName for the computer
#                          Note this option can be set multiple times
#                          If it is not set, default servicePrincipalNames are specified for the computer
# -enctype <enc>
#               (optional) Specified msDS-SupportedEncryptionTypes for the computer
#                          Note this option can be set multiple times
#                          If it is not set, default msDS-SupportedEncryptionTypes are specified for the computer
# -licensetype <license type>
#               (optional) Specified which license type to use for the precreate computer.
#                          It can be "server" or "workstation".
#                          If empty, default to "server".
#                          
# Example:  Precreate computer "redhat" at default zone in centrify.com 
#           select_zone {CN=default,CN=Zones,DC=centrify,DC=com}
#           precreate_computer redhat$@centrify.com -trustee foo@centrify.com -trustee bar@child.centrify.com
proc precreate_computer {upn args} {
    set usage "precreate_computer samaccount@domain"
    set usage "$usage ?-ad? ?-scp? ?-nogc? ?-czone? ?-all? ?-container rdn? ?-dnsname dnsname? ?-trustee upn1? ?-trustee upn2? ?-stype spn1? ?-stype spn2? ?-enctype enc1? ?-enctype enc2? ?-licensetype licensetype? ...";

    # Resolve samaccount@domain
    set dotIndex [string last "$@" $upn]
    set len [string length $upn]
    if {$dotIndex <= 0 || $dotIndex >= [expr $len-2]} {
        error "Computer name should be: samaccount@domain, and the samaccount should end with $"
    }
    set computer [string range $upn 0 [expr $dotIndex-1]]
    set domain [string range $upn [expr $dotIndex+2] $len]

    set base [dn_from_domain $domain]
    set sam "$computer\$"
    set pw [string tolower [string range $computer 0 13]]

    set createAD ""
    set createScp ""
    set noGC ""
    set createZone ""
    set explicitlyCreateAD ""
    set explicitlyCreateZone ""
    set trustees ""
    set container ""
    set dnsname ""
    set stypes ""
    set enctypes ""
    # adedit does not support auto zone, so, in adedit, all zone's default license type are "server"
    set licensetype "server"
    set argCount [llength $args]
    for {set idx 0} {$idx < $argCount} {incr idx} {
        set flag [lindex $args $idx]
        switch -glob -- $flag {
            -ad {
               set createAD "true" 
               set explicitlyCreateAD "true"
            }
            -scp {
               set createScp "true" 
            }
            -nogc {
               set noGC "true"
            }
            -czone {
               set createZone "true" 
               set explicitlyCreateZone "true"
            }
            -all {
               set createAD "true" 
               set createScp "true" 
               set createZone "true" 
            }
            -container {
                incr idx
                if {$idx >= $argCount} {
                    error "Missing -container argument: should be\n$usage"
                }
                set container [lindex $args $idx] 
            }
            -dnsname {
                incr idx
                if {$idx >= $argCount} {
                    error "Missing -dnsname argument: should be\n$usage"
                }
                set dnsname [lindex $args $idx] 
            }
            -trustee {
                incr idx
                if {$idx >= $argCount} {
                    error "Missing -trustee argument: should be\n$usage"
                }
                lappend trustees [lindex $args $idx] 
            }
            -stype {
                incr idx
                if {$idx >= $argCount} {
                    error "Missing -stype argument: should be\n$usage"
                }
                lappend stypes [lindex $args $idx] 
            }
            -enctype {
                incr idx
                if {$idx >= $argCount} {
                    error "Missing -enctype argument: should be\n$usage"
                }
                lappend enctypes [lindex $args $idx] 
            }
            -licensetype {
                incr idx
                if {$idx >= $argCount} {
                    error "Missing -licensetype argument: should be\n$usage"
                }
                set licensetype [lindex $args $idx]
                if {$licensetype != "server" && $licensetype != "workstation"} {
                    error "Unknown license type: should be \"server\" or \"workstation\""
                }
            }
            -* {
                error "Unknown flag $flag: should be\n$usage"
            }
            default {
                error "Unknown option $flag: should be\n$usage"
            }
        }
    }

    if {$createAD != "true" && $createScp != "true" && $createZone != "true"} {
        # Default to create ad and scp
        set createAD "true" 
        set createScp "true" 
    }

    set ztype ""
    set zone ""
    set zbase ""
    if {$createScp == "true" || $createZone == "true"} {
        set ztype [gzf type]
        if {$ztype != "tree" && $ztype != "classic3" &&  $ztype != "classic4"} {
            error "Wrong zone type $ztype, should be tree zone or classic zone"
        }
        set zone [gzf dn]
        set zbase [dn_from_domain [domain_from_dn $zone]]
    }
    if {$dnsname == ""} {
        # If not specified, dnsname should be computer.<joined domain>(same as console)
        set dnsname $computer.$domain
    }

    set addn ""
    if {$createAD == "true" || $createScp == "true"} {
        set objlist [get_objects -depth sub $base "(&(objectclass=computer)(cn=$computer))"]
        if {[llength $objlist] <= 0 && $createAD == "true"} {
            # If ad computer object does not exist and createAD
            # Create ad computer
            if {[string length $container] == 0} {
                set container computers
            }
            if {![catch {select_object $container}]} {
                set addn cn=$computer,$container
            } elseif {![catch {select_object $container,$base}]} {
                set addn cn=$computer,$container,$base
            } elseif {![catch {select_object cn=$container,$base}]} {
                set addn cn=$computer,cn=$container,$base
            } else {
                error "Invalid container"
            }

            new_object $addn
            sof objectclass computer 
            sof samaccountname $sam
            # Do not enable the computer object here.
            sof useraccountcontrol [expr 0x1002]
            svo
        } elseif {[llength $objlist] <= 0 && $createAD != "true"} { 
            # If ad computer object does not exist and not createAD
            # Error
            error "AD computer object doesn't exist, please specify -ad to create"
        } elseif {$explicitlyCreateAD == "true"} { 
            # If ad computer object exists
            # only error when explicitly creating AD computer object
            error "AD computer object already exists"
        } else {
            set addn [lindex $objlist 0]
        }

        # If computer object is enabled, process will continue. This behavior
        # is consistent with windows console.
        # So no need to check if the computer object is enabled or not here.

        # If need to create SCP, ensure that there's no existent scp among the
        # forest (REF: #54307). Otherwise, exit before any object is created.
        if {$createScp == "true"} {
            select_object $addn; 
            set sid [get_object_field sid]
            set forest [get_bind_info $domain forest]
            set filter "(|(keywords=parentLink:$sid)(managedBy=$addn))";
            if {$noGC == "true"} {
                set scpdn [get_objects -depth sub $base $filter]
            } else {
                set scpdn [get_objects -gc -f $forest -depth sub "" $filter]
            }
            if {[llength $scpdn] > 0} {
                # If we previously created a scp before, will encounter this
                # error too.
                error "AD computer object is joined or previously joined to zone. Please use another unused computer object."
            }
        }

        # Change computer password
        sup "$sam@$domain" $pw

        # Enable computer account
        select_object $addn
        set control [gof userAccountControl]
        set control [expr $control & ~0x2]
        sof useraccountcontrol $control
        sof dNSHostName $dnsname

        set newPostalAddress {}
        foreach postalAddressItem [gof postalAddress] {
            if {[string first "LicenseType:" $postalAddressItem] == -1} {
                lappend newPostalAddress $postalAddressItem
            }
        }
        lappend newPostalAddress "LicenseType:$licensetype"
        sof postalAddress {*}$newPostalAddress
        # When option -stype was not specified, command will apply with default
        # SPN list.
        # Note: need to be synchronized to Direct Access Manage default SPN list
        if {[llength $stypes] == 0} {
            set stypes "afpserver cifs ftp host http ipp nfs"
        }
        set spns ""
        # Guarantee every SPN is unique.
        foreach stype [lsort -unique $stypes] {
            lappend spns $stype/$computer
            lappend spns $stype/$dnsname
        }
        eval "sof servicePrincipalName $spns"
        
        # When option -enctype was not specified, command will apply with default
        # encryption list.
        if {[llength $enctypes] == 0} {
            set enctypes "aes256-cts aes128-cts arcfour-hmac-md5 des-cbc-md5 des-cbc-crc"
        }
        set enc 0
        foreach enctype [lsort -unique $enctypes] {
            switch -regexp -- $enctype {
                aes256-cts-hmac-sha1-96|aes256-cts {
                  set enc [expr $enc | 16]
                }
                aes128-cts-hmac-sha1-96|aes128-cts {
                  set enc [expr $enc | 8]
                }
                arcfour-hmac|rc4-hmac|arcfour-hmac-md5 {
                  set enc [expr $enc | 4]
                }
                des-cbc-md5|^des$ {
                  set enc [expr $enc | 2]
                }
                des-cbc-crc {
                  set enc [expr $enc | 1]
                }
            }
        }
        sof msDS-SupportedEncryptionTypes $enc
        svo
    }

    # Create service connection point object.
    if {$createScp == "true"} {
        new_zone_computer $sam@$domain
        set_zone_computer_field enabled 1
        save_zone_computer

        select_object $addn; 
        set scpdn [get_objects -depth sub $zbase "(|(keywords=parentLink:[gof sid])(managedBy=$addn))"]
        if {[llength $scpdn] != 1} {
            error "Can not find the scp we just created, maybe replication problem."
        } else {
            set scpdn [lindex $scpdn 0]
        }

        set uacGuid [get_schema_guid User-Account-Control] 
        # Well-known User-Force-Change-Password
        set pwdGuid {00299570-246d-11d0-a768-00aa006e0529}
        # Well-known Validated-DNS-Host-Name
        set dhnGuid {72e39547-7b18-11d1-adef-00c04fd8d5cd}
        # Well-known Validated-SPN
        set spnGuid {f3a64788-5306-11d1-a9c5-0000f80367c1}
        foreach trustee $trustees {
            slo [principal_to_dn $trustee]
            set usid [gof sid]
            
            slo $scpdn
            set sd [gof sd]
            # Arbitrary Trustee, Zone/Computers/scp, This object, Generic Read
            set sd [add_sd_ace $sd "A;;GR;;;$usid"]
            sof sd $sd
            svo

            slo $addn
            set sd [gof sd]
            # Arbitrary Trustee, AD computer object, This object, This object, Reset Password
            set sd [add_sd_ace $sd "OA;;CR;$pwdGuid;;$usid"]
            # Arbitrary Trustee, AD computer object, This object, This object, Write userAccountControl
            set sd [add_sd_ace $sd "OA;;RPWP;$uacGuid;;$usid"]
            # Arbitrary Trustee, AD computer object, This object, This object, Validated write to DNS host name 
            set sd [add_sd_ace $sd "OA;;SWRP;$dhnGuid;;$usid"]
            # Arbitrary Trustee, AD computer object, This object, This object, read/write to service principal name
            set sd [add_sd_ace $sd "OA;;RPWP;$spnGuid;;$usid"]
            sof sd $sd
            svo
        }

        set osGuid [get_schema_guid Operating-System] 
        set osvGuid [get_schema_guid Operating-System-Version]
        set oshGuid [get_schema_guid Operating-System-Hotfix]
        set osspGuid [get_schema_guid Operating-System-Service-Pack]
        slo $addn
        set csid [gof sid]
        set sd [gof sd]
        # SELF, AD computer object, This object, Write operatingSystem
        set sd [add_sd_ace $sd "OA;;WP;$osGuid;;PS"]
        # SELF, AD computer object, This object, Write operatingSystemVersion
        set sd [add_sd_ace $sd "OA;;WP;$osvGuid;;PS"]
        # SELF, AD computer object, This object, Write operatingSystemHotfix
        set sd [add_sd_ace $sd "OA;;WP;$oshGuid;;PS"]
        # SELF, AD computer object, This object, Write operatingSystemServicePack
        set sd [add_sd_ace $sd "OA;;WP;$osspGuid;;PS"]
        # SELF, AD computer object, This object, Reset Password
        set sd [add_sd_ace $sd "OA;;CR;$pwdGuid;;PS"]
        # SELF, AD computer object, This object, read userAccountControl
        set sd [add_sd_ace $sd "OA;;RP;$uacGuid;;PS"]
        # SELF, AD computer object, This object, Validated write to DNS host name 
        set sd [add_sd_ace $sd "OA;;SWRP;$dhnGuid;;PS"]
        # SELF, AD computer object, This object, read/write to service principal name
        set sd [add_sd_ace $sd "OA;;RPWP;$spnGuid;;PS"]        
        sof sd $sd
        svo

        set kwdGuid [get_schema_guid Keywords] 
        set disGuid [get_schema_guid Display-Name]
        slo $scpdn
        set sd [gof sd]
        # Precreated computer account, Zone/Computer/scp, This object, Generic Read
        set sd [add_sd_ace $sd "A;;GR;;;$csid"]
        # Precreated computer account, Zone/Computer/scp, This object, Write keywords property
        set sd [add_sd_ace $sd "OA;;WP;$kwdGuid;;$csid"]
        # Precreated computer account, Zone/Computer/scp, This object, Write displayName property
        set sd [add_sd_ace $sd "OA;;WP;$disGuid;;$csid"]
        sof sd $sd
        svo
    }

    if {$createZone == "true"} {
        set MZ $dnsname@$zone
        catch {select_zone $MZ;dlz}
        set czone_type [expr { $ztype == "tree" ? "computer" : "classic-computer" }]
        create_zone $czone_type $MZ

        # restore current zone
        select_zone $zone
    }
}

proc getopt {_argv name {_var ""} } {
    upvar 1 $_argv argv $_var var
    set pos [lsearch -regexp $argv ^$name]
    if {$pos>=0} {
        set to $pos
        if {$_var ne ""} {
            set var [lindex $argv [incr to]]
        }
        set argv [lreplace $argv $pos $to]
        return 1
    } else {
        return 0
    }
}

proc convert_msdate {msdate} {
    if {$msdate==9223372036854775807} {
        return -1
    }
    return [clock format [expr ($msdate/10000000)-11644473600]]
}

set dict_gt [dict create]
dict set dict_gt 0x0001 "SYSTEM" 
dict set dict_gt 0x0002 "GLOBAL" 
dict set dict_gt 0x0004 "DOMAIN_LOCAL" 
dict set dict_gt 0x0008 "UNIVERSAL" 
dict set dict_gt 0x0010 "APP_BASIC" 
dict set dict_gt 0x0020 "APP_QUEEY" 
dict set dict_gt 0x80000000 "SECURITY" 

proc explain_groupType {gt} {
    global dict_gt
    set r [format "%x" $gt]
    dict for {k v} $dict_gt {
        if [expr $gt&$k] {
            lappend r $v
	    }
    }
    return $r
}

set dict_ta [dict create]
dict set dict_ta 0x0001 "NON_TRANSITIVE" 
dict set dict_ta 0x0002 "UPLEVEL_ONLY" 
dict set dict_ta 0x0004 "QUARANTINED_DOMAIN" 
dict set dict_ta 0x0008 "FOREST_TRANSITIVE" 
dict set dict_ta 0x0010 "CROSS_ORGANIZATION" 
dict set dict_ta 0x0020 "WITHIN_FOREST" 
dict set dict_ta 0x0040 "TREAT_AS_EXTERNAL" 
dict set dict_ta 0x0080 "USE_RC4_ENCRYPTION" 

proc explain_trustAttributes {ta} {
    global dict_ta
    set r [format "%x" $ta]
    dict for {k v} $dict_ta {
        if [expr $ta&$k] {
            lappend r $v
        }
    }
    return $r
}

proc explain_trustDirection {td} {
    switch $td {
	0 {return "disabled"}
	1 {return "inbound"}
	2 {return "outbound"}
	3 {return "two-way"}
        default {return "unknown"}
    }
}

set dict_pt [dict create]
dict set dict_pt # "Local Unix User"
dict set dict_pt % "Local Unix Group"
dict set dict_pt $ "Local Windows User"
dict set dict_pt : "Local Windows Group"
dict set dict_pt a "All AD users"
dict set dict_pt x "All Unix users"
dict set dict_pt w "All Windows users"
dict set dict_pt u "AD user"
dict set dict_pt g "AD group"

proc explain_ptype {pt} {
    global dict_pt
     if {![dict exists $dict_pt $pt]} {
        error "invalid ptype"
    } else {
        return [dict get $dict_pt $pt]
    }
}

set dict_uac [dict create]
dict set dict_uac 0x0001 "ADS_UF_SCRIPT" 
dict set dict_uac 0x0002 "ADS_UF_ACCOUNTDISABLE" 
dict set dict_uac 0x0008 "ADS_UF_HOMEDIR_REAUIRED" 
dict set dict_uac 0x0010 "ADS_UF_LOCKOUT" 
dict set dict_uac 0x0020 "ADS_UF_PASSWD_BITREQD" 
dict set dict_uac 0x0040 "ADS_UF_PASSWD_CANT_CHANGE" 
dict set dict_uac 0x0080 "ADS_UF_ENCRYPTED_TEXT_PASSWORD_ALLOWED" 
dict set dict_uac 0x0100 "ADS_UF_TEMP_DUPLICATE_ACCOUNT" 
dict set dict_uac 0x0200 "ADS_UF_NORMAL_ACCOUNT" 
dict set dict_uac 0x0800 "ADS_UF_INTERDOMAIN_TRUST_ACCOUNT" 
dict set dict_uac 0x1000 "ADS_UF_WORKSTATION_TRUST_ACCOUNT" 
dict set dict_uac 0x2000 "ADS_UF_SERVER_TRUST_ACCOUNT" 
dict set dict_uac 0x010000 "ADS_UF_DONT_EXPIRE_PASSWD" 
dict set dict_uac 0x020000 "ADS_UF_MNS_LOGON_ACCOUNT" 
dict set dict_uac 0x040000 "ADS_UF_SMARTCARD_REQUIRED" 
dict set dict_uac 0x080000 "ADS_UF_TRUSTED_FOR_DELEGATION" 
dict set dict_uac 0x100000 "ADS_UF_NOT_DELEGATED" 
dict set dict_uac 0x200000 "ADS_UF_USE_DES_KEY_ONLY" 
dict set dict_uac 0x400000 "ADS_UF_DONT_REQUIRE_PREAUTH" 
dict set dict_uac 0x800000 "ADS_UF_PASSWORD_EXPIRED" 

proc explain_userAccountControl {uac} {
    global dict_uac
    set r [format "%x" $uac]
    dict for {k v} $dict_uac {
        if [expr $uac&$k] {
            lappend r $v
	    }
    }
    return $r
}

proc list_zones {domain} {
    push
    foreach  zone [get_zones $domain] {
	if {[catch {select_zone -nc $zone}]} {
            puts "Cannot open zone: $zone"
        } else {
            puts "$zone : [gzf type] : [gzf schema] " }
    }
    pop
}

# Modify a timebox string (42 hex characters)
# strTimeBox:   original timebox string(42 hex characters)
# day:          the day of week to modify(Sunday=0)
# hour:         the hour to modify(0-23)
#               0 means 0:00~1:00, 23 means 23:00~24:00
# avail:        0 - make this hour unavailable
#               not 0 - make this hour available
# Example:  Let Saturday 23:00~24:00(Sa: 23:00-00:00) available
#           set tb 000000000000000000000000000000000000000000
#           puts [modify_timebox $tb 6 23 1]
#           800000000000000000000000000000000000000000
proc modify_timebox {strTimeBox day hour avail} {
    # Trim the head 0 in hour
    set hour [string trimleft $hour 0]
    if {$hour == ""} {
        set hour 0
    }
    if {$hour < 0 || $hour > 23} {
        error "Invalid hour: $hour. Should be 0~23"
    }
    if {$day < 0 || $day > 6} {
        error "Invalid day of week: $day. Should be 0~6"
    }
    set dayIndex [expr 6*int($day)+2]
    # 4 bit a hex characters, each bit represents one hour
    set hourIndex [expr int($hour/4)]
    # Because the timebox string is little-ending, and the higher bit is bigger hour
    # So change 0 1 2 3 4 5 -> 1 0 3 2 5 4
    set hourIndex [expr int($hourIndex/2)*4+1-$hourIndex] 
    set hourMask [expr int(pow(2, int($hour)%4))]
    # calculate index char 
    set i [expr ($dayIndex+$hourIndex)%42]
    # get original value
    set v 0x[string index $strTimeBox $i]
    # modify value
    if {$avail == "0"} {
        set v [format "%X" [expr $v & ~$hourMask]]
    } else {
        set v [format "%X" [expr $v | $hourMask]]
    }
    return [string replace $strTimeBox $i $i $v]
}

####################################################################
# timebox: a string(42 hex characters)    
#          Each day takes up 6 hex characters
#   internal format: the return value of [get_role_field timebox] 
#                    start from Sat 16:00 and the order is unusual 
#   output formal: an usual expression begin at Sun 0:00,and in usual order 
# internal format:             [2*i+1]         [2*i]           --- string index
#     [ 0~ 1] Sat  16:00~24:00 [ 1]16:00~20:00 [ 0]20:00~24:00
#     [ 2~ 3] Sun  00:00~08:00 [ 3]00:00~04:00 [ 2]04:00~08:00
#     [ 4~ 5]      08:00~16:00 [ 5]08:00~12:00 [ 4]12:00~16:00
#     [ 6~ 7]      16:00~24:00 [ 7]16:00~20:00 [ 6]16:00~24:00
#     [ 8~13] Mon  00:00~24:00        ......
#     ......     ......               ......
#     [32~37] Fri  00:00~24:00        ......
#     [38~39] Sat  00:00~08:00 [39]00:00~04:00 [38]04:00~08:00
#     [40~41] Sat  08:00~16:00 [41]08:00~12:00 [40]12:00~16:00
# output format:               [2*i]           [2*i+1]         --- string index
#     [ 0~ 1] Sun  00:00~08:00 [ 0]00:00~04:00 [ 1]04:00~08:00
#     [ 2~ 3]      08:00~16:00 [ 2]08:00~12:00 [ 3]12:00~16:00
#     [ 4~ 5]      16:00~24:00 [ 4]16:00~20:00 [ 5]16:00~24:00
#     [ 6~11] Mon  00:00~24:00        ......
#     ......     ......               ......
#     [30~35] Fri  00:00~24:00        ......
#     [36~37] Sat  00:00~08:00 [36]00:00~04:00 [37]04:00~08:00
#     [38~39] Sat  08:00~16:00 [38]08:00~12:00 [39]12:00~16:00
#     [40~41] Sat  16:00~24:00 [40]16:00~20:00 [41]20:00~24:00
#####################################################################
# return output format
proc decode_timebox {internal} {
    set output $internal
    # move the last 8 hours of Sat--the first 2 characters--to last,
    # the order is remaining
    set Sat_20_24_hours [string index $output 0]
    set Sat_16_20_hours [string index $output 1]
    set output [string replace $output 0 1]
    append output $Sat_20_24_hours 
    append output $Sat_16_20_hours 
    # now output start from 0:00 of Sun.
    # exchange value at position of 2i and 2i+1 hex characters .
    # make the expression of 8 hours is 
    #  (0:00-4:00 4:00-8:00) (8:00-12:00 12:00-16:00) (16:00-20:00 20:00-24:00)
    for {set i 0} {$i <21} {incr i} {
        # exchange the hex characters of the 2 position 
        set high_pos  [expr $i+$i] 
        set high [string index $output $high_pos] 
        set low_pos   [expr $high_pos+1] 
        set low  [string index $output $low_pos ] 
        set output [string replace $output $high_pos $high_pos $low  ]
        set output [string replace $output $low_pos  $low_pos  $high ]
    }
    return $output 
}
# return internal format
proc encode_timebox {output} {
    set internal $output 
    # move the last 8 hours of Sat--the first 2 characters-- to first,
    # the order is remaining
    set Sat_20_24_hours [string index $internal 41]
    set Sat_16_20_hours [string index $internal 40]
    set internal [string replace $internal 40 41]
    set Sat_16_24_hours ""
    append Sat_16_beginning $Sat_16_20_hours 
    append Sat_16_beginning $Sat_20_24_hours 
    append Sat_16_beginning $internal 
    set internal $Sat_16_beginning 
    # now internal start from 16:00 of Sat
    # exchange value at position of 2i and 2i+1 hex characters.
    # make the expression of 8 hours is 
    #  (4:00-8:0 00:00-4:00) (12:00-16:00 8:00-12:00) (20:00-24:00 16:00-20:00)
    for {set i 0} {$i <21} {incr i} {
        # exchange the hex characters of the 2 position 
        set high_pos  [expr $i+$i] 
        set high [string index $internal $high_pos] 
        set low_pos   [expr $high_pos+1] 
        set low  [string index $internal $low_pos ] 
        set internal [string replace $internal $high_pos $high_pos $low  ]
        set internal [string replace $internal $low_pos  $low_pos  $high ]
    }
    return $internal 
}

# args: rolename, description, sysrights,pamapplist,dzcmdlist,allowlocaluser,rse,visible
proc create_role {role {desc {}} {sr {}} {apps {}} {cmds {}} {allowlocal {}} {rse {}} {visible {}} } {
    new_role $role
    
    set_role_field description $desc
	
    if {[string length $allowlocal]} {
        srf allowlocaluser $allowlocal
    }
	
    if {[string length $sr]} {
        srf sysrights $sr
    }
    
    if {[string length $visible]} {
        srf visible $visible
    }

    foreach app $apps {
        add_pamapp_to_role $app
    }
    foreach cmd $cmds {
        add_command_to_role $cmd
    }
    if {[string length $rse]} {
        set_rs_env_for_role $rse
    }
    save_role
}

# args: pamappname, application, description
proc create_pam_app {pam app {desc {}}} {
    new_pam_app $pam;
    set_pam_field application $app;
    spf description $desc;
    save_pam_app;
}
# args: dzcmdname, cmd, description,form,
# dzdo_runas,dzsh_runas,flags,priority,umask,path,
# selinux_role,selinux_type.
proc create_dz_command {dzc cmd 
                        {desc {}} 
                        {form {}} 
                        {dzdo_runas {}} 
                        {dzsh_runas {}} 
                        {flags {}} 
                        {pri {}} 
                        {umask {}}
                        {path {}}
                        {selinux_role {}}
                        {selinux_type {}}} {
    new_dz_command $dzc 
    set_dzc_field cmd $cmd
    sdzcf description $desc
    sdzcf dzdo_runas $dzdo_runas
    if {[string length $dzsh_runas]} {
        sdzcf dzsh_runas $dzsh_runas
    }
    if {[string length $form]} {
        sdzcf form $form
    }
    if {[string length $flags]} {
        sdzcf flags  $flags
    }
    if {[string length $pri]} {
        sdzcf pri $pri
    }
    if {[string length $umask]} {
        sdzcf umask $umask
    }
    if {[string length $path]} {
        sdzcf path  $path
    }
    if {[string length $selinux_role]} {
        sdzcf selinux_role  $selinux_role
    }
    if {[string length $selinux_type]} {
        sdzcf selinux_type  $selinux_type
    }
    save_dz_command
}

# args: rscmdname, cmd, description,form,
# dzsh_runas,flags,priority,umask,path.
proc create_rs_command {rsc cmd 
                        {desc {}} 
                        {form {}} 
                        {dzsh_runas {}} 
                        {flags {}} 
                        {pri {}} 
                        {umask {}}
                        {path {}}} {
    new_rs_command $rsc 
    set_rsc_field cmd $cmd
    srscf description $desc
    srscf dzsh_runas $dzsh_runas
    if {[string length $form]} {
        srscf form $form
    }
    if {[string length $flags]} {
        srscf flags  $flags
    }
    if {[string length $pri]} {
        srscf pri $pri
    }
    if {[string length $umask]} {
        srscf umask $umask
    }
    if {[string length $path]} {
        srscf path  $path
    }
    save_rs_command
}

# args: rsename, description
proc create_rs_env {rse {desc {}}} {
    new_rs_env $rse
    set_rse_field description $desc
    save_rs_env
}

# args: userdn allow
proc modify_change_pwd {userdn allow} {
    slo $userdn
    set pwguid "ab721a53-1e2f-11d0-9819-00aa0040529b"
    set sd [gof sd]
    set sdlist [split $sd ()]
    set i 0
    set newsd ""
    set flag ""
    set acetype ""

    foreach ace $sdlist {
        incr i
        if {$i == 1} {
            set flag $ace
            continue
        }

        if {$ace == ""} {
            continue
        }

        lassign [split $ace {;}] - - - objecttype - trustee
        if {($trustee == "PS" || $trustee == "WD") && $objecttype == $pwguid} {
            continue 
        } else { 
            set newsd "${newsd}($ace)"
        }
    }

    if {$allow == "true"} {
        set acetype "OA"
    } else {
        set acetype "OD"
    }

    set newsd "${flag}($acetype;;CR;$pwguid;;PS)($acetype;;CR;$pwguid;;WD)$newsd"
    sof sd $newsd
    save_object
}

# args: userdn
proc set_change_pwd_denied {userdn} {
    modify_change_pwd $userdn false
}

# args: userdn
proc set_change_pwd_allowed {userdn} {
    modify_change_pwd $userdn true
}

# go back along the zone hierarchy from current level to base
# to get the attribute value from user's unix profile.
#
# args: upn     user's upn
#       attr    attribute name
# return    attribute value if found
#           empty if not
proc get_user_unix_profile { upn attr } {
    push
    set ret ""
    set ztype [get_zone_field type]
    if {$ztype == "tree" || $ztype == "classic3"
        || $ztype == "classic4" || $ztype == "computer"} {
        set zone [gzf dn]
    } else {
        error "Wrong zone type $ztype, should be tree zone, classic zone or computer zone"
    }
    while { $zone != "" } {
        if {![catch {slzu $upn}]} {
            set ret [gzuf $attr]
            if { $ret == "-" } {
                set ret ""
            }
            if { $ret != "" } {
                break
            }
        }

        if { [catch {set zone [gzf parent]}] } {
            break
        }
        if {$zone == ""} {
            break
        }
        slz $zone
    }

    pop
    return $ret
}

# go back along the zone hierarchy from current level to base
# to get the attribute value from group's unix profile.
# args: upn     group's upn
#       attr    attribute name
# return    attribute value if found
#           empty if not
proc get_group_unix_profile { upn attr } {
    push
    set ret ""
    set ztype [gzf type]
    if {$ztype == "tree" || $ztype == "classic3"
        || $ztype == "classic4" || $ztype == "computer"} {
        set zone [gzf dn]
    } else {
        error "Wrong zone type $ztype, should be tree zone, classic zone or computer zone"
    }
    while { $zone != "" } {
        if {![catch {slzg $upn}]} {
            set ret [gzgf $attr]
            if { $ret == "-" } {
                set ret ""
            }
            if { $ret != "" } {
                break
            }
        }

        if { [catch {set zone [gzf parent]}] } {
            break
        }
        if {$zone == ""} {
            break
        }
        slz $zone
    }
    pop
    return $ret
}

# get all of the AD users along zone hierarchy from current zone to root
# This function supports only tree zone and classic zone.
# For classic zone, our search always starts from current zone.
# For tree zone, our search starts from computer zone, if "-hostname" is given.
#
# arguments:
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#                               If it is set, the search starts from computer zone (if existed),
#                               If not, the search starts from current zone.
#
#   return                      a list of AD user dn
#
proc get_all_zone_users {args} {
    push
    set usage {wrong # args: should be [-upn] [-hostname <host name>]}

    # get arguments
    set getupn [getopt args "-upn"]
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    if { [llength $args] > 0 } {
        error $usage
    }
    set zoneDN [get_zone_field dn]

    set options ""
    if { $getupn } {
        set options "-upn"
    }

    set ADusers [dict create]

    # decide where our search starts
    set ztype [get_zone_field type]
    if {$ztype == "classic3" || $ztype == "classic4"} {
        foreach user [gzu {*}$options] {
            dict set ADusers [string tolower $user] $user
        }
    } elseif {$ztype == "tree"} {
        if { $doComputerZone } {
            set zone $hostname@$zoneDN
            if { [catch {slz $zone}] } {
                # computer zone may not exist, skip it
                set zone $zoneDN
                set doComputerZone 0
            }
        } else {
            set zone $zoneDN
        }

        # go along zone hierarchy from current zone to root
        while { $zone != "" } {
            slz $zone

            foreach user [gzu {*}$options] {
                dict set ADusers [string tolower $user] $user
            }

            if { [catch {set zone [gzf parent]}] } {
                break
            }
        }
    } else {
        error "Wrong zone type $ztype, should be tree zone or classic zone"
    }

    pop
    return [dict values $ADusers]
}

# get all of the AD groups along zone hierarchy from current zone to root
# This function supports only tree zone and classic zone.
# For classic zone, our search always starts from current zone.
# For tree zone, our search starts from computer zone, if "-hostname" is given.
#
# arguments:
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#                               If it is set, the search starts from computer zone (if existed);
#                               If not, the search starts from current zone.
#
#   return                      a list of AD groups dn
#
proc get_all_zone_groups {args} {
    push
    set usage {wrong # args: should be [-upn] [-hostname <host name>]}

    # get arguments
    set getupn [getopt args "-upn"]
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    if { [llength $args] > 0 } {
        error $usage
    }
    set zoneDN [get_zone_field dn]

    set options ""
    if { $getupn } {
        set options "-upn"
    }

    # decide where our search starts
    set ztype [get_zone_field type]

    if {$ztype == "classic3" || $ztype == "classic4"} {
        foreach group [gzg {*}$options] {
            dict set ADgroups [string tolower $group] $group
        }
    } elseif {$ztype == "tree"} {
        if { $doComputerZone } {
            set zone $hostname@$zoneDN
            if { [catch {slz $zone}] } {
                # computer zone may not exist, skip it
                set zone $zoneDN
                set doComputerZone 0
            }
        } else {
            set zone $zoneDN
        }

        set ADgroups {}
        # go along zone hierarchy from current zone to root
        while { $zone != "" } {
            slz $zone

            foreach group [gzg {*}$options] {
                dict set ADgroups [string tolower $group] $group
            }

            if { [catch {set zone [gzf parent]}] } {
                break
            }
        }
    } else {
        error "Wrong zone type $ztype, should be tree zone or classic zone"
    }

    pop
    return [dict values $ADgroups]
}

# select an AD object
# return all of groups, which it belongs to.
#
# arguments:
#   $target object DN
#   return  a list of memberships
#
proc get_all_membership { target } {
    push
    slo $target
    set domain [dn_from_domain [domain_from_dn $target]]

    set memberships {}
    # get all global and universal groups
    foreach sid [sid_to_ms_string [gof "tokenGroupsGlobalAndUniversal"]] {
        set upn [principal_from_sid $sid]
        set dn [principal_to_dn $upn]
        lappend memberships $dn
    }

    # get local groups
    set clist [concat [list $target] $memberships]
    while { [llength $clist] != 0 } {
        set nlist {}
        foreach dn $clist {
            # groupType = ADS_GROUP_TYPE_SECURITY_ENABLED | ADS_GROUP_TYPE_LOCAL_GROUP
            #           = 0x80000000 | 0x00000004
            set gt [expr 0x80000004]
            set sid [principal_to_sid [dn_to_principal $dn]]
            foreach ent [get_objects -depth sub "$domain" "(&(objectCategory=group)(member=<sid=$sid>)(groupType=$gt))"] {
                if {[lsearch $memberships $ent] == -1} {
                    lappend memberships $ent
                    lappend nlist $ent
                }
            }
        }
        set clist $nlist
    }

    pop
    return $memberships
}

# collect the computer role assignments for a specific computer
# It checks the target group of each assignment, and only takes
# the effective assignments into account.
# We say an assignment is effective, only if the computer is subjective to it.
#
# arguments:
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#   -visible                    return only the role assignments which have visible roles
#
#   return                      a list of role assignments
#
proc get_computer_role_assignments { args } {
    push
    set usage {wrong # args: should be [-visible] [-hostname <host name>]}
    
    # get arguments
    set checkVisible [getopt args "-visible"]
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    if { [llength $args] > 0 } {
        error $usage
    }

    set visibleOption {}
    if { $checkVisible } {
        set visibleOption "-visible"
    }

    set assignments {}

    # Split the fqdn hostname into host and domain
    set domain [split $hostname {.}]
    set host [lindex $domain 0]
    set domain [join [lrange $domain 1 end] {.}]

    # check the membership of the computer
    set hostDN [principal_to_dn "$host\$@$domain"]
    set memberShips [get_all_membership $hostDN]

    if { [catch {set cRoles [get_child_zones -crole]}] } {
        # have no computer role in current zone
        pop
        return $assignments
    }

    # collect computer role
    foreach cRole $cRoles {
        slz $cRole

        set found 0
        set groups [gzf computers]
        foreach group $groups {
            if {[catch {set group [principal_to_dn $group]}]} {
                # the group may not exist
                continue
            }
            if { [lsearch -nocase $memberShips $group] >= 0 } {
                set found 1
                break
            }
        }
        if {!$found} {
            # go ahead to next computer role
            continue
        }

        # collect the role assignees
        set assignments [concat $assignments [get_role_assignments {*}$visibleOption]]
    }
    pop
    return $assignments
}

# check whether a group is an effective zone group
# This function supports only tree zone and classic zone.
# For classic zone, our search always starts from current zone.
# For tree zone, our search starts from computer zone, if "-hostname" is given.
#
# arguments: should be [-hostname <host name>] [-unix] <name>
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#                               If it is set, the search starts from computer zone (if existed);
#                               If not, the search starts from current zone.
#
#   -unix                       If set, we locate the group by its unix name;
#                               If not, we locate it by its upn
#
#   -case                       If set, do case sensitive compare to local and
#                               AD names.
#
#   name                        group name
#
#   return                      return 1 if it is an effective group
#
proc is_group_effective { args } {
    push
    set usage {wrong # args: should be [-hostname <host name>] [-unix] <name>}

    # get arguments
    set byUnixName [getopt args "-unix"]
    set doCase [getopt args "-case"]
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    set name {}
    if { [llength $args] == 1 } {
        lassign $args name
    } else {
        error $usage
    }

    set zoneDN [get_zone_field dn]
    set ztype [get_zone_field type]

    if {$ztype == "classic3" || $ztype == "classic4"} {
        set found 0
        # check if it exists
        catch {
            if { $byUnixName } {
                slzg -unix $name
            } else {
                slzg $name
            }
            set found 1
        }
        if { !$found } {
            # it does not exist
            pop
            return 0
        }
    } elseif {$ztype == "tree"} {
        # decide where our search starts
        if { $doComputerZone } {
            set zone $hostname@$zoneDN
            if { [catch {slz $zone}] } {
                # computer zone may not exist, skip it
                set zone $zoneDN
                set doComputerZone 0
            }
        } else {
            set zone $zoneDN
        }

        # trace along the zone hierarchy
        set zones {}
        while { $zone != "" } {
            slz $zone
            lappend zones $zone
            if { [catch {set zone [gzf parent]}] } {
                break
            }
        }

        # check if the group exists
        # if exist, get its upn
        set found 0
        set zone_level -1
        foreach zone $zones {
            incr zone_level
            slz $zone
            if { $byUnixName } {
                if {[catch {slzg -unix $name}]} {
                    continue
                }
                set uname [gzgf name]
                # check if it is a fake entry
                # get its upn
                if { [catch {set upn [dn_to_principal [gzgf addn]]}] } {
                    # cannot get upn, that means it has been deleted
                    # or it is from one-way trust domain
                    # get its SID, then continue
                    foreach group [gzg] {
                        # not all groups from gzg can be selected
                        catch {
                            if {[string index $group 0] == "*"} {
                                slzg $group
                                if { [gzgf name] == $name } {
                                    set upn $group
                                    break
                                }
                            }
                        }
                    }
                }

                # check if upn is overrided with different name in lower zone level
                set beOverrided 0
                while { $zone_level > 0} {
                    incr zone_level -1
                    slz [lindex $zones $zone_level]
                    if {![catch {slzg $upn}]} {
                        set unix_name [gzgf name]
                        if {$unix_name != "" && $unix_name != "-"} {
                            if {$doCase} {
                                if {$unix_name != $name} {
                                    set beOverrided 1
                                    break
                                }
                            } else {
                                if {[string compare -nocase $unix_name $name] != 0} {
                                    set beOverrided 1
                                    break
                                }
                            }
                        }
                    }
                }
                if { $beOverrided } {
                    continue
                }
                set found 1
            } else {
                if {[catch {slzg $name}]} {
                    continue
                }
                set uname [gzgf name]
                set upn [dn_to_principal [gzgf addn]]
                set found 1
            }

            if { $found } {
                # compare the name we found
                if {$doCase} {
                    if {$uname eq $name} {
                        break
                    }
                } else {
                    if {[string compare -nocase $uname $name] == 0} {
                        break
                    }
                }
                set found 0
            }
        }
        if { !$found } {
            # $name does not exist
            pop
            return 0
        }

        # check if the unix profile is complete
        set found 1
        set attrWanted {gid name}
        # do we start from computer zone ?
        if { $doComputerZone } {
            set zone $hostname@$zoneDN
        } else {
            set zone $zoneDN
        }
        slz $zone
        foreach attr $attrWanted {
            set val [get_group_unix_profile $upn $attr]
            if { $val == "" } {
                set found 0
                break
            }
        }
        if { !$found } {
            # the unix profile is not complete
            pop
            return 0
        }
    } else {
        error "Wrong zone type $ztype, should be tree zone or classic zone"
    }

    pop
    return 1
}

# check whether a user is an effective zone user
# This function supports only tree zone and classic zone.
# For classic zone, our search always starts from current zone.
# For tree zone, our search starts from computer zone, if "-hostname" is given.
#
# arguments:
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#                               If it is set, the search starts from computer zone (if existed),
#                               and computer role is taken into account;
#                               If not, the search starts from current zone,
#                               and computer role is ignored
#
#   -unix                       If set, we locate the user by its unix name;
#                               If not, we locate it by its upn
#
#   -case                       If set, do case sensitive compare to local and
#                               AD names.
#
#   -dzcache                    If set, look up user's effective sysrights from
#                               DZ cache instead of walking through the zone
#                               tree and look for role assignments.
#                               Note that the DZ cache file must be opened
#                               beforehand.
#
#   -name                       user name
#
#   return                      return 1 if it is an effective user
#
proc is_user_effective { args } {
    push
    set usage {wrong # args: should be [-hostname <host name>] [-dzcache] [-unix] <name>}

    # get arguments
    set byUnixName [getopt args "-unix"]
    set doCase [getopt args "-case"]
    set doDZCache [getopt args "-dzcache"]
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    set name {}
    if { [llength $args] == 1 } {
        lassign $args name
    } else {
        error $usage
    }

    set zoneDN [get_zone_field dn]
    set ztype [get_zone_field type]

    if { $doDZCache } {
        if { ![is_dzcache_opened] } {
            error "No DZ cache is opened"
        }
    }

    if {$ztype == "classic3" || $ztype == "classic4"} {
        set found 0
        catch {
            # check if it exists
            if { $byUnixName } {
                slzu -unix $name
            } else {
                slzu $name
            }
            # check if it is enabled
            if { [gzuf enabled] } {
                set found 1
            }
        }
        if { !$found } {
            # it does not exist
            pop
            return 0
        }
    } elseif {$ztype == "tree"} {
        # which zone do we start from
        if { $doComputerZone } {
            set zone $hostname@$zoneDN
            if { [catch {slz $zone}] } {
                # computer zone may not exist, skip it
                set zone $zoneDN
                set doComputerZone 0
            }
        } else {
            set zone $zoneDN
        }

        # trace along the zone hierarchy
        set zones {}
        while { $zone != "" } {
            slz $zone
            lappend zones $zone
            if { [catch {set zone [gzf parent]}] } {
                break
            }
        }

        # check if the user exists
        # if exist, get its upn
        set found 0
        set zone_level 0
        foreach zone $zones {
            incr zone_level
            slz $zone
            if { $byUnixName } {
                if {[catch {slzu -unix $name}]} {
                    continue
                }
                set uname [gzuf uname]
                # check if it is a fake entry
                # get its upn
                if { [catch {set upn [dn_to_principal [gzuf addn]]}] } {
                    # cannot get upn, that means it has been deleted
                    # or it is from one-way trust domain
                    # get its SID, then continue
                    # if it has role assigned, we think it is effective
                    foreach user [gzu] {
                        # not all users from gzu can be selected
                        catch {
                            if {[string index $user 0] == "*"} {
                                slzu $user
                                if { [gzuf uname] == $name } {
                                    set upn $user
                                    break
                                }
                            }
                        }
                    }
                }

                #check if upn is overrided with different name in lower zone level
                set beOverrided 0
                while { $zone_level > 0} {
                    incr zone_level -1
                    slz [lindex $zones $zone_level]
                    if {![catch {slzu $upn} ]} {
                        set unix_name [gzuf uname]
                        if {$unix_name != "" && $unix_name != "-"} {
                            if {$doCase} {
                                if {$unix_name != $name} {
                                    set beOverrided 1
                                    break
                                }
                            } else {
                                if {[string compare -nocase $unix_name $name] != 0} {
                                    set beOverrided 1
                                    break
                                }
                            }
                        }
                    }
                }
                if { $beOverrided } {
                    continue
                }
                set found 1
            } else {
                if {[catch {slzu $name}]} {
                    continue
                }
                set uname [gzuf uname]
                set upn [dn_to_principal [gzuf addn]]
                set found 1
            }

            if { $found } {
                # compare the name we found
                if {$doCase} {
                    if {$uname eq $name} {
                        break
                    }
                } else {
                    if {[string compare -nocase $uname $name] == 0} {
                        break
                    }
                }
                set found 0
            }
        }
        if { !$found } {
            # $name does not exist
            pop
            return 0
        }

        # check if the unix profile is complete
        set found 1
        set attrWanted {uid gid shell home uname}
        # do we start from computer zone ?
        if { $doComputerZone } {
            set zone $hostname@$zoneDN
        } else {
            set zone $zoneDN
        }
        slz $zone
        foreach attr $attrWanted {
            set val [get_user_unix_profile $upn $attr]
            if { $val == "" } {
                set found 0
                break
            }
        }
        if { !$found } {
            # the unix profile is not complete
            pop
            return 0
        }

        set dict_mem [dict create]
        # the upn of duplicate entry has "+1" at end
        # remove it, get real upn
        set real_upn $upn
        regexp {(.*)\+[0-9]+$} $upn - real_upn

        set found 0
        if { !$doDZCache } {
            # check if have role assigned
            foreach zone $zones {
                catch {
                    slz $zone
                    # collect all role assignments by zone levels
                    set assignments [get_role_assignments -visible]
                    if { [info exists hostname] } {
                        set assignments [concat $assignments [get_computer_role_assignments -visible -hostname $hostname]]
                    }

                    # check the assignee
                    foreach assignment $assignments {
                        set items [split $assignment "/"]
                        set assignee [lindex $items 0]
                        if { $assignee == "all ad users" || $assignee == $real_upn } {
                            set found 1
                        } elseif { [string index $upn 0] != "*" } {
                            # get membership
                            # it may be costly, so keep it in dict for next use
                            if {[dict exists $dict_mem $real_upn]} {
                                set memberships [dict get $dict_mem $real_upn]
                            } else {
                                # upn may be badname, need to catch it.
                                # for detail, please try
                                # principal_to_dn badname@smoke2k8.centrify.test
                                set memberships {}
                                catch {set memberships [get_all_membership [principal_to_dn $real_upn]]}
                                dict set dict_mem $real_upn $memberships
                            }
                            foreach group $memberships {
                                set group [dn_to_principal $group]
                                if { $assignee == $group } {
                                    set found 1
                                    break
                                }
                            }
                        }
                        if { $found } {
                            break
                        }
                    }
                    # This breaks out of the catch.
                    if { $found } {
                        break
                    }
                }
                # This breaks out of the zone loop.
                if { $found } {
                    break
                }
            }
        } else {
            # query DZ cache database
            set user_dn [principal_to_dn $real_upn]
            if { [string index $upn 0] != "*" } {
                # get membership
                # it may be costly, so keep it in dict for next use
                if {[dict exists $dict_mem $real_upn]} {
                    set memberships [dict get $dict_mem $real_upn]
                } else {
                    # upn may be badname, need to catch it.
                    # for detail, please try
                    # principal_to_dn badname@smoke2k8.centrify.test
                    set memberships {}
                    catch {set memberships [get_all_membership $user_dn]}
                    dict set dict_mem $real_upn $memberships
                }
                set groups {}
                foreach group $memberships {
                    set sid [principal_to_sid [dn_to_principal $group]]
                    lappend groups $sid
                }
                if {![catch {lassign [read_dzcache_user_sysrights $user_dn {*}$groups] sysrights visible}]} {
                    set found $visible
                }
            }
        }
        if { !$found } {
            # have no role assigned for $name
            pop
            return 0
        }
    }

    pop
    return 1
}

# collect the role assignments along zone hierarchy from current zone to root
# This function supports only tree zone and classic zone.
# For classic zone, our search always starts from current zone.
# For tree zone, our search starts from computer zone, if "-hostname" is given.
#
# arguments:
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#                               If it is set, the search starts from computer zone (if existed),
#                               and computer role is taken into account;
#                               If not, the search starts from current zone,
#                               and computer role is ignored
#
#   -visible                    return only the role assignments which have visible roles
#
#   return                      a list of AD groups dn
#
proc get_all_zone_role_assignments { args } {
    push
    set usage {wrong # args: should be [-visible] [-hostname <host name>]}

    # get arguments
    set checkVisible [getopt args "-visible"]
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    if { [llength $args] > 0 } {
        error $usage
    }

    set options {}
    if { $doComputerZone } {
        lappend options "-hostname" $hostname
    }

    set visibleOption {}
    if { $checkVisible } {
        set visibleOption "-visible"
    }

    set zoneDN [get_zone_field dn]
    set ztype [get_zone_field type]
    if {$ztype != "tree" && $ztype != "classic3" && $ztype != "classic4"} {
        error "Wrong zone type $ztype, should be tree zone or classic zone"
    }

    # decide where our search starts
    if { $doComputerZone } {
        set zone $hostname@$zoneDN
        if { [catch {slz $zone}] } {
            # computer zone may not exist, skip it
            set zone $zoneDN
            set doComputerZone 0
        }
    } else {
        set zone $zoneDN
    }

    # go through the zone tree from child to parent
    set assignments {}
    while { $zone != "" } {
        slz $zone
        set assignments [concat $assignments [get_role_assignments {*}$visibleOption]]

        if { [gzf type] == "tree" && [info exists hostname] } {
            set assignments [concat $assignments [get_computer_role_assignments {*}$visibleOption {*}$options]]
        }

        if { [catch {set zone [gzf parent]}] } {
            break
        }
    }

    pop
    return $assignments
}

# collect the users along zone hierarchy from current zone to root.
# The users whose unix profile is incomplete are filtered out.
# The users who have not any role assigned are filtered out.
# This function supports only tree zone and classic zone.
# For classic zone, our search always starts from current zone.
# For tree zone, our search starts from computer zone, if "-hostname" is given.
#
# arguments:
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#                               If it is set, the search starts from computer zone (if existed),
#                               and computer role is taken into account;
#                               If not, the search starts from current zone,
#                               and computer role is ignored
#
#   return                      a list of unix name
#
proc get_effective_users { args } {
    push
    set usage {wrong # args: should be [-hostname <host name>]}

    # get arguments
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    if { [llength $args] > 0 } {
        error $usage
    }
    set zoneDN [get_zone_field dn]

    set options {}
    if { $doComputerZone } {
        lappend options "-hostname" $hostname
    }

    set visibleOption "-visible"

    set effectiveUsers {}
    # get all zone users
    # we may get *SID here, it may be caused by
    # 1, the user was deleted, it will be discarded later
    #   when find it has no role assigned.
    # 2, it is from a one-way trust domain
    #   we continue to check its scp and role
    set ADusers [get_all_zone_users {*}$options]

    set ztype [get_zone_field type]
    if {$ztype == "classic3" || $ztype == "classic4"} {
        foreach upn $ADusers {
            if { [get_user_unix_profile $upn enabled] } {
                lappend effectiveUsers [get_user_unix_profile $upn uname]
            }
        }
    } elseif {$ztype == "tree"} {
        set assignments [get_all_zone_role_assignments {*}$visibleOption {*}$options]

        # collect assignees
        foreach assignment $assignments {
            set items [split $assignment "/"]
            set assignees([lindex $items 0]) 1
        }

        # filter out the in-effective users/groups who have not any role assigned
        set attrWanted {uid gid shell home uname}
        if { $doComputerZone } {
            set zone $hostname@$zoneDN
            if { [catch {slz $zone}] } {
                # computer zone may not exist, skip it
                slz $zoneDN
                set doComputerZone 0
            }
        }

        foreach upn $ADusers {
            set doNext 0
            foreach attr $attrWanted {
                if { [get_user_unix_profile $upn $attr] == ""} {
                    set doNext 1
                    break
                }
            }
            if { $doNext } {
                # it is in-effective, its unix profile is incomplete
                continue
            }

            # the upn of duplicate entry has "+1" at end
            # remove it, get real upn
            set real_upn $upn
            regexp {(.*)\+[0-9]+$} $upn - real_upn
            # check role assignment
            set doNext 1
            if { [info exists "assignees(all ad users)"]
                || [info exists assignees($real_upn)] } {
                set doNext 0
            } elseif { [string index $upn 0] != "*" } {
                foreach group [get_all_membership [principal_to_dn $real_upn]] {
                    set group [dn_to_principal $group]
                    if { [info exists assignees($group)] } {
                        set doNext 0
                        break
                    }
                }
            }
            if { $doNext } {
                # it is in-effective, has no any role assigned.
                continue
            }

            lappend effectiveUsers [get_user_unix_profile $upn uname]
        }
    } else {
        error "Wrong zone type $ztype, should be tree zone or classic zone"
    }

    pop
    return $effectiveUsers
}

# collect the groups along zone hierarchy from current zone to root.
# The groups whose unix profile is incomplete are filtered out.
# This function supports only tree zone and classic zone.
# For classic zone, our search always starts from current zone.
# For tree zone, our search starts from computer zone, if "-hostname" is given.
#
# arguments:
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#                               If it is set, the search starts from computer zone (if existed);
#                               If not, the search starts from current zone.
#
#   return                      a list of unix name
#
proc get_effective_groups { args } {
    push
    set usage {wrong # args: should be [-hostname <host name>]}

    # get arguments
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    if { [llength $args] > 0 } {
        error $usage
    }
    set zoneDN [get_zone_field dn]

    set options {}
    if { $doComputerZone } {
        lappend options "-hostname" $hostname
    }

    set effectiveGroups {}
    set ADgroups {}
    foreach group [get_all_zone_groups {*}$options] {
        # ignore those groups whose name starts with "*"
        if { [string index $group 0] != "*" } {
            lappend ADgroups $group
        }
    }

    set ztype [get_zone_field type]
    if {$ztype == "classic3" || $ztype == "classic4"} {
        foreach upn $ADgroups {
            lappend effectiveGroups [get_group_unix_profile $upn name]
        }
    } elseif {$ztype == "tree"} {
        # filter out the in-effective group who has no complete unix profile
        set attrWanted {gid name}
        if { $doComputerZone } {
            set zone $hostname@$zoneDN
            if { [catch {slz $zone}] } {
                # computer zone may not exist, skip it
                slz $zoneDN
                set doComputerZone 0
            }
        }

        foreach upn $ADgroups {
            set doNext 0
            foreach attr $attrWanted {
                if { [get_group_unix_profile $upn $attr] == ""} {
                    set doNext 1
                    break
                }
            }
            if { $doNext } {
                # it is in-effective, its unix profile is incomplete
                continue
            }

            lappend effectiveGroups [get_group_unix_profile $upn name]
        }
    } else {
        error "Wrong zone type $ztype, should be tree zone or classic zone"
    }

    pop
    return $effectiveGroups
}

# compare the contents of two files
# Refer to: http://wiki.tcl.tk/4808
# args:
#   file1   the file path
#   file2   the path of the other file
#   return 0, if they are totally consistent to each other
#   return non-zero, otherwise
#   return error if file does not exist
proc compare_files {file1 file2} {
    set s1 [file size $file1]
    set s2 [file size $file2]
    set diff 0
    if {$s1 == $s2} {
        if {[catch {set f1 [open $file1 r]} errmsg]} {
            error "cannot open $file1: $errmsg"
        }
        if {[catch {set f2 [open $file2 r]} errmsg]} {
            error "cannot open $file2: $errmsg"
        }
        fconfigure $f1 -translation binary
        fconfigure $f2 -translation binary
        while {![eof $f1] || ![eof $f2]} {
            set d1 [read $f1 2048]
            set d2 [read $f2 2048]
            set diff [string compare $d1 $d2]
            if {$diff != 0} {
                break
            }
        }
        close $f1
        close $f2
    } else {
        set diff 1
    }
    return $diff
}

# Collect role assignments in the current zone for the given user.
# The wanted role assignments are given based on AD group membership 
# or directly assigned to AD user.
# Find the AD user, resolve the group that the AD user is a member of, 
# and locate all the role assignments including computer roles 
# for the current zone.
#
# arguments:
#   dn                          the dn of AD user
#   -hostname <host name>       the fqdn of computer, can get it by "adinfo host"
#                               If it is set, we will search computer role 
#                               under current zone.
#                               If not, we ignore computer role.
#
#   -visible                    return only the role assignments which have visible roles
#
#   return                      a list of role assignments
#
proc get_user_role_assignments {args} {
    push
    set usage {wrong # args: should be [-visible] [-hostname <host name>] <dn>}
    # get arguments
    set checkVisible [getopt args "-visible"]
    set hostname {}
    set doComputerZone [getopt args "-hostname" hostname]
    set dn {}
    if { [llength $args] == 1 } {
        lassign $args dn
    } else {
        error $usage
    }

    set options {}
    if { $doComputerZone } {
        lappend options "-hostname" $hostname
    }

    set visibleOption {}
    if { $checkVisible } {
        set visibleOption "-visible"
    }

    set upn [dn_to_principal $dn]

    # collect all groups which the user belongs to
    set dict_mem [dict create]
    foreach group [get_all_membership $dn] {
        set group [dn_to_principal $group]
        dict set dict_mem $group "1"
    }

    # collect all role assignments in current zone
    set assignments [gra {*}$visibleOption]
    if { [info exists hostname] } {
        # catch it if computer role does not exist
        catch {
            set assignments [concat $assignments [get_computer_role_assignments {*}$visibleOption {*}$options]]
        }
    }

    set ret {}
    # check the assignee
    foreach assignment $assignments {
        set items [split $assignment "/"]
        set assignee [lindex $items 0]
        if { $assignee == "all ad users" || $assignee == $upn } {
            lappend ret $assignment
        } else {
            if {[dict exists $dict_mem $assignee]} {
                lappend ret $assignment
            }
        }
    }

    pop
    return $ret
}

proc load_properties { args } {

    global properties

    # Clear the dictionary
    set properties {} 

    # Load properties
    set file {/etc/centrifydc/centrifydc.conf}
    if { [catch {set conf [open $file r]} errmsg] } {
        error "Cannot open file $file: $errmsg"
    }
    while {[gets $conf line] != -1} {
        set line [string trim $line]
        if {$line eq "" || [string first "#" $line] == 0} {
            continue
        }
        if {[regexp {^\s*([[:alnum:]\._-]+)(\s*:\s*|\s*=\s*|\s*:=\s*|\s+)(.*)} $line - prop sep val]} {
            dict set properties $prop $val
        }
    }
}

proc list_properties { args } {

    global properties

    if {[info exists properties]} {
        foreach name [lsort [dict keys $properties]] {
            puts "$name: \"[dict get $properties $name]\""
        }
    }
}

proc get_property { args } {

    set usage "get_property propertyname"
    set usage "$usage ?-default defaultvalue?";
    
    # Get arguments
    set defval {}
    getopt args "-default" defval
    set name {}
    if {[llength $args] == 1} {
        lassign $args name
    } else {
        error $usage
    }

    global properties

    # Load properties for the first time
    if {![info exists properties]} {
        load_properties
    }

    if {[dict exists $properties $name]} {
        return [dict get $properties $name]
    } else {
        return $defval
    }
}
