#!/bin/sh /usr/share/centrifydc/perl/run.cda

##############################################################################
#
# Copyright (C) 2014-2019 Centrify Corporation. All rights reserved.
#
# We need ourselves "flock" to prevent conflictions between multiple threads.
# Because the target file may be a symlink,
# and the file may be closed or unlinked before all jobs were done.
#
##############################################################################

package autoedit_file_lock;

BEGIN
{
    use lib "/usr/share/centrifydc/scripts/da";
    use lib "/usr/share/centrifydc/perl";
}

use strict;
use warnings;
use CentrifyDA::Lock;

sub new
{
    if (scalar(@_) == 0)
    {
        return undef;
    }

    my $this = {};
    my $path = shift;
    my $FH = undef;

    # if it is a WPAR system, and the file is a symlink under /usr,
    # we lock the real file
    my $lswpar = "/usr/sbin/lswpar";
    if (-x $lswpar)
    {
        if (system("$lswpar 2> /dev/null") && $path =~ /^\/*usr\//)
        {
            my $real = readlink($path);
            if ($real)
            {
                $path = $real;
            }
        }
    }

    $path .= ".cdc.lock";

    my $count = 10;
    while(--$count)
    {
        $FH = CentrifyDA::Lock->new($path, 1);
        last if defined $FH;
        sleep(1);
    }

    if ($count <= 0)
    {
        return undef;
    }

    $$this{"path"} = $path;
    $$this{"lock"} = $FH;
    
    bless $this;
    return $this;
}

1;
